"""The `__about__` module exposes the version of the `nats_tools` package:

Example:

```python
from nats_tools.__about__ import __version__
print(__version__)
```

Note that container version can also be imported directly from `nats_tools` package:

Example:

```python
from nats_tools import __version__
print(__version__)
```
"""
__version__ = "1.0.1"
